@@ -1,8 +1,7 @@
 # -*- coding: utf-8; mode: tcl; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- vim:fenc=utf-8:ft=tcl:et:sw=4:ts=4:sts=4
 # portuninstall.tcl
-# $Id$
 #
-# Copyright (c) 2004-2005, 2008-2011 The MacPorts Project
+# Copyright (c) 2004-2005, 2008-2011, 2014-2015 The MacPorts Project
 # Copyright (c) 2002 - 2003 Apple Inc.
 # All rights reserved.
 #
@@ -41,6 +40,64 @@ set UI_PREFIX "---> "
 
 namespace eval registry_uninstall {
 
+# generate list of all dependencies of the port
+proc generate_deplist {port {optslist ""}} {
+
+    set deptypes {depends_fetch depends_extract depends_build depends_lib depends_run depends_test}
+    set all_dependencies {}
+    # look up deps from the saved portfile if possible
+    if {![catch {set mport [mportopen_installed [$port name] [$port version] [$port revision] [$port variants] $optslist]}]} {
+        array set depportinfo [mportinfo $mport]
+        mportclose $mport
+        foreach type $deptypes {
+            if {[info exists depportinfo($type)]} {
+                foreach dep $depportinfo($type) {
+                    lappend all_dependencies [lindex [split $dep :] end]
+                }
+            }
+        }
+        # append those from the registry (could be different because of path deps)
+        foreach dep [$port dependencies] {
+            lappend all_dependencies [$dep name]
+        }
+    } else {
+        # grab the deps from the dep map
+        set portname [$port name]
+        set depmaplist [registry::list_depends $portname [$port version] [$port revision] [$port variants]]
+        foreach dep $depmaplist {
+            lappend all_dependencies [lindex $dep 0]
+        }
+        # and the ones from the current portfile
+        if {![catch {mportlookup $portname} result] && [llength $result] >= 2} {
+            array set depportinfo [lindex $result 1]
+            set porturl $depportinfo(porturl)
+            set variations {}
+            set minusvariant [lrange [split [registry::property_retrieve $port negated_variants] -] 1 end]
+            set plusvariant [lrange [split [$port variants] +] 1 end]
+            foreach v $plusvariant {
+                lappend variations $v "+"
+            }
+            foreach v $minusvariant {
+                lappend variations $v "-"
+            }
+            if {![catch {set mport [mportopen $porturl [concat $optslist subport $portname] [array get variations]]} result]} {
+                array unset depportinfo
+                array set depportinfo [mportinfo $mport]
+                mportclose $mport
+            }
+            foreach type $deptypes {
+                if {[info exists depportinfo($type)]} {
+                    foreach dep $depportinfo($type) {
+                        lappend all_dependencies [lindex [split $dep :] end]
+                    }
+                }
+            }
+        }
+    }
+    set all_dependencies [lsort -unique $all_dependencies]
+    return $all_dependencies
+}
+
 # takes a composite version spec rather than separate version,revision,variants
 proc uninstall_composite {portname {v ""} {optionslist ""}} {
     if {$v eq ""} {
@@ -52,8 +109,7 @@ proc uninstall_composite {portname {v ""} {optionslist ""}} {
 }
 
 proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""}} {
-    global uninstall.force uninstall.nochecksum UI_PREFIX \
-           macports::portimagefilepath macports::registry.path
+    global uninstall.force UI_PREFIX macports::registry.path
     array set options $optionslist
     if {[info exists options(subport)]} {
         # don't want this set when calling registry::run_target
@@ -85,18 +141,34 @@ proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""
             append composite_spec _${revision}${variants}
         }
     }
-    set ilist [eval registry::entry imaged $searchkeys]
+    set ilist [registry::entry imaged {*}$searchkeys]
     if { [llength $ilist] > 1 } {
         # set portname again since the one we were passed may not have had the correct case
         set portname [[lindex $ilist 0] name]
-        ui_msg "$UI_PREFIX [msgcat::mc "The following versions of $portname are currently installed:"]"
-        foreach i [portlist_sortint $ilist] {
-            set ispec "[$i version]_[$i revision][$i variants]"
+        set msg "The following versions of $portname are currently installed:"
+        if {[macports::ui_isset ports_noninteractive]} {
+            ui_msg "$UI_PREFIX [msgcat::mc $msg]"
+        }
+        set sortedlist [portlist_sortint $ilist]
+        foreach i $sortedlist {
+            set portstr [format "%s @%s_%s%s" [$i name] [$i version] [$i revision] [$i variants]]
             if {[$i state] eq "installed"} {
-                ui_msg "$UI_PREFIX [format [msgcat::mc "    %s @%s (active)"] [$i name] $ispec]"
-            } else {
-                ui_msg "$UI_PREFIX [format [msgcat::mc "    %s @%s"] [$i name] $ispec]"
+                append portstr [msgcat::mc " (active)"]
             }
+
+            if {[info exists macports::ui_options(questions_multichoice)]} {
+                lappend portilist "$portstr"
+            } else {
+                ui_msg "$UI_PREFIX     $portstr"
+            }
+        }
+        if {[info exists macports::ui_options(questions_multichoice)]} {
+            set retstring [$macports::ui_options(questions_multichoice) $msg "Choice_Q2" $portilist]
+            foreach index $retstring {
+                set uport [lindex $sortedlist $index]
+                uninstall [$uport name] [$uport version] [$uport revision] [$uport variants]
+            }
+            return 0
         }
         throw registry::invalid "Registry error: Please specify the full version as recorded in the port registry."
     } elseif { [llength $ilist] == 1 } {
@@ -112,6 +184,7 @@ proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""
         throw registry::invalid "Registry error: ${portname}${composite_spec} not registered as installed"
     }
 
+    set userinput {}
     # uninstall dependents if requested
     if {[info exists options(ports_uninstall_follow-dependents)] && $options(ports_uninstall_follow-dependents) eq "yes"} {
         # don't uninstall dependents' dependencies
@@ -134,83 +207,46 @@ proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""
         }
     } else {
         # check its dependents
-        registry::check_dependents $port ${uninstall.force} "uninstall"
+        set userinput [registry::check_dependents $port ${uninstall.force} "uninstall"]
+        if {$userinput eq "quit"} {
+            return 0
+        }
     }
     # if it's active, deactivate it
     if {[$port state] eq "installed"} {
         if {[info exists options(ports_dryrun)] && [string is true -strict $options(ports_dryrun)]} {
             ui_msg "For $portname @${composite_spec}: skipping deactivate (dry run)"
         } else {
-            if {[info exists options(ports_uninstall_no-exec)] || ![registry::run_target $port deactivate $optionslist]} {
-                portimage::deactivate $portname $version $revision $variants [array get options]
+            if {$userinput eq "forcedbyuser"} {
+                set options(ports_nodepcheck) "yes"
+            }
+            if {[info exists options(ports_uninstall_no-exec)] || ![registry::run_target $port deactivate [array get options]]} {
+                if {$userinput eq "forcedbyuser"} {
+                    portimage::deactivate $portname $version $revision $variants [array get options]
+                    unset options(ports_nodepcheck) 
+                } else {
+                    portimage::deactivate $portname $version $revision $variants [array get options]
+                }
             }
         }
     }
 
-    set ref $port
-
-    # note deps before we uninstall if we're going to uninstall them too
+    # note deps before we uninstall if we're going to uninstall them too (i.e. --follow-dependencies)
     if {[info exists options(ports_uninstall_follow-dependencies)] && [string is true -strict $options(ports_uninstall_follow-dependencies)]} {
-        set deptypes {depends_fetch depends_extract depends_build depends_lib depends_run}
-        set all_dependencies {}
-        # look up deps from the saved portfile if possible
-        if {![catch {set mport [mportopen_installed [$port name] [$port version] [$port revision] [$port variants] $optionslist]}]} {
-            array set depportinfo [mportinfo $mport]
-            mportclose $mport
-            foreach type $deptypes {
-                if {[info exists depportinfo($type)]} {
-                    foreach dep $depportinfo($type) {
-                        lappend all_dependencies [lindex [split $dep :] end]
-                    }
-                }
-            }
-            # append those from the registry (could be different because of path deps)
-            foreach dep [$port dependencies] {
-                lappend all_dependencies [$dep name]
-            }
-        } else {
-            # grab the deps from the dep map
-            set depmaplist [registry::list_depends $portname $version $revision $variants]
-            foreach dep $depmaplist {
-                lappend all_dependencies [lindex $dep 0]
-            }
-            # and the ones from the current portfile
-            if {![catch {mportlookup $portname} result] && [llength $result] >= 2} {
-                array set depportinfo [lindex $result 1]
-                set porturl $depportinfo(porturl)
-                set variations {}
-                set minusvariant [lrange [split [registry::property_retrieve $ref negated_variants] -] 1 end]
-                set plusvariant [lrange [split $variants +] 1 end]
-                foreach v $plusvariant {
-                    lappend variations $v "+"
-                }
-                foreach v $minusvariant {
-                    lappend variations $v "-"
-                }
-                if {![catch {set mport [mportopen $porturl [concat $optionslist subport $portname] [array get variations]]} result]} {
-                    array unset depportinfo
-                    array set depportinfo [mportinfo $mport]
-                    mportclose $mport
-                }
-                foreach type $deptypes {
-                    if {[info exists depportinfo($type)]} {
-                        foreach dep $depportinfo($type) {
-                            lappend all_dependencies [lindex [split $dep :] end]
-                        }
-                    }
-                }
-            }
-        }
-        array unset depportinfo
-        set all_dependencies [lsort -unique $all_dependencies]
+        set all_dependencies [registry_uninstall::generate_deplist $port $optionslist]
     }
 
     if {[info exists options(ports_dryrun)] && [string is true -strict $options(ports_dryrun)]} {
         ui_msg "For $portname @${composite_spec}: skipping uninstall (dry run)"
+        # allow deps to not be excluded from the list below just because this port is still a dependent
+        if {[info exists options(ports_uninstall_follow-dependencies)] && [string is true -strict $options(ports_uninstall_follow-dependencies)]} {
+            set uports [list [list $portname $version $revision $variants]]
+        }
     } else {
         ui_msg "$UI_PREFIX [format [msgcat::mc "Uninstalling %s @%s"] $portname $composite_spec]"
 
         # Get the full path to the image file
+        set ref $port
         set imagefile [registry::property_retrieve $ref location]
         file delete $imagefile
         # Try to delete the port's image dir; will fail if there are more image
@@ -231,7 +267,7 @@ proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""
         }
 
         set portfile_path [file join ${registry.path} registry portfiles ${portname}-${version}_${revision} $portfile]
-        if {[registry::entry search portfile $portfile] eq {}} {
+        if {[registry::entry search portfile $portfile name $portname version $version revision $revision] eq {}} {
             file delete -force $portfile_path
             catch {file delete [file dirname $portfile_path]}
         }
@@ -249,44 +285,76 @@ proc uninstall {portname {version ""} {revision ""} {variants 0} {optionslist ""
             }
         }
     }
-    
-    # uninstall dependencies if requested
+
+    if {![info exists uports]} {
+        set uports {}
+    }
+    # create list of all dependencies that will be uninstalled, if requested
     if {[info exists options(ports_uninstall_follow-dependencies)] && [string is true -strict $options(ports_uninstall_follow-dependencies)]} {
-        # don't uninstall dependencies' dependents
-        if {[info exists options(ports_uninstall_follow-dependents)]} {
-            unset options(ports_uninstall_follow-dependents)
-            set optionslist [array get options]
-        }
-        while 1 {
-            set remaining_list {}
-            foreach dep $all_dependencies {
-                if {![catch {set ilist [registry::installed $dep]}]} {
-                    set remaining 0
-                    foreach i $ilist {
-                        set iversion [lindex $i 1]
-                        set irevision [lindex $i 2]
-                        set ivariants [lindex $i 3]
-                        if {[llength [registry::list_dependents $dep $iversion $irevision $ivariants]] == 0} {
-                            set regref [registry::open_entry $dep $iversion $irevision $ivariants [lindex $i 5]]
-                            if {![registry::property_retrieve $regref requested] && ([info exists options(ports_uninstall_no-exec)] || ![registry::run_target $regref uninstall $optionslist])} {
-                                registry_uninstall::uninstall $dep $iversion $irevision $ivariants $optionslist
-                            }
-                        } else {
-                            set remaining 1
-                        }
+        set alldeps $all_dependencies
+        set portilist {}
+        for {set j 0} {$j < [llength $alldeps]} {incr j} {
+            set dep [lindex $alldeps $j]
+            set uninstalling_this_dep 0
+            if {![catch {set ilist [registry::installed $dep]}]} {
+                foreach i $ilist {
+                    lassign $i dep iversion irevision ivariants
+                    if {[list $dep $iversion $irevision $ivariants] in $uports} {
+                        continue
                     }
-                    if {$remaining} {
-                        lappend remaining_list $dep
+                    set regref [registry::open_entry $dep $iversion $irevision $ivariants [lindex $i 5]]
+                    if {![registry::property_retrieve $regref requested]} {
+                        set all_dependents_uninstalling 1
+                        foreach depdt [$regref dependents] {
+                            if {[list [$depdt name] [$depdt version] [$depdt revision] [$depdt variants]] ni $uports} {
+                                set all_dependents_uninstalling 0
+                                break
+                            }
+                        }
+                        if {$all_dependents_uninstalling} {
+                            lappend uports [list $dep $iversion $irevision $ivariants]
+                            lappend portilist $dep@${iversion}_${irevision}${ivariants}
+                            set uninstalling_this_dep 1
+                        }
                     }
                 }
             }
-            if {[llength $remaining_list] == 0 || [llength $remaining_list] == [llength $all_dependencies]} {
-                break
+            if {$uninstalling_this_dep} {
+                set deprefs [registry::entry imaged $dep]
+                foreach depref $deprefs {
+                    set depdeps [registry_uninstall::generate_deplist $depref $optionslist]
+                    foreach d $depdeps {
+                        if {$d ni [lrange $alldeps $j+1 end]} {
+                            lappend alldeps $d 
+                        }
+                    }
+                }
+            }
+        }
+        ## User Interaction Question
+        # show a list of all dependencies to be uninstalled with a timeout when --follow-dependencies is specified
+        if {[info exists macports::ui_options(questions_yesno)] && [llength $uports] > 0 && !([info exists options(ports_dryrun)] && [string is true -strict $options(ports_dryrun)])} {
+            $macports::ui_options(questions_yesno) "The following dependencies will be uninstalled:" "Timeout_1" $portilist {y} 10
+        }
+        unset options(ports_uninstall_follow-dependencies)
+    }
+
+    # uninstall all dependencies in order from uports
+    foreach dp $uports {
+        lassign $dp iname iversion irevision ivariants
+        if {![catch {registry::open_entry $iname $iversion $irevision $ivariants ""} regref]} {
+            if {[info exists options(ports_dryrun)] && [string is true -strict $options(ports_dryrun)]} {
+                if {$iname ne $portname} {
+                    ui_msg "For $iname @${iversion}_${irevision}${ivariants}: skipping uninstall (dry run)"
+                }
+            } else {
+                if {[info exists options(ports_uninstall_no-exec)] || ![registry::run_target $regref uninstall [array get options]]} {
+                    registry_uninstall::uninstall $iname $iversion $irevision $ivariants [array get options]
+                }
             }
-            set all_dependencies $remaining_list
         }
     }
-    
+
     return 0
 }
 
